// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: sourceforge.net/projects/libusbdotnet/
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Runtime.InteropServices;
using LibUsbDotNet.Internal;

#pragma warning disable 649

namespace LibUsbDotNet.Descriptors
{
    /// <summary> Usb Endpoint Descriptor
    /// </summary> 
    /// <remarks> This is the actual descriptor as described in the USB 2.0 Specifications.
    /// </remarks> 
    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public class UsbEndpointDescriptor : UsbDescriptor
    {
        /// <summary>
        /// Total size of this structure in bytes.
        /// </summary>
        public new static readonly int Size = Marshal.SizeOf(typeof (UsbEndpointDescriptor));

        /// <summary>
        /// Endpoint Address
        /// Bits 0..3b Endpoint Number.
        /// Bits 4..6b Reserved. Set to Zero
        /// Bits 7 Direction 0 = Out, 1 = In (Ignored for Control Endpoints)
        /// </summary>
        public readonly byte EndpointID;

        /// <summary>
        /// Bits 0..1 Transfer Type 
        /// 00 = Control
        /// 01 = Isochronous
        /// 10 = Bulk
        /// 11 = Interrupt
        /// 
        /// Bits 2..7 are reserved. If Isochronous endpoint, 
        /// Bits 3..2 = Synchronisation Type (Iso Mode) 
        /// 00 = No Synchonisation
        /// 01 = Asynchronous
        /// 10 = Adaptive
        /// 11 = Synchronous
        /// 
        /// Bits 5..4 = Usage Type (Iso Mode) 
        /// 00 = Data Endpoint
        /// 01 = Feedback Endpoint
        /// 10 = Explicit Feedback Data Endpoint
        /// 11 = Reserved
        /// </summary>
        public readonly byte Attributes;

        /// <summary>
        /// Maximum Packet Size this endpoint is capable of sending or receiving
        /// </summary>
        public readonly short MaxPacketSize;

        /// <summary>
        /// Interval for polling endpoint data transfers. Value in frame counts. Ignored for Bulk and Control Endpoints. Isochronous must equal 1 and field may range from 1 to 255 for interrupt endpoints.
        /// </summary>
        public readonly byte Interval;

        /// <summary>
        /// Audio endpoint specific.
        /// </summary>
        public readonly byte Refresh;

        /// <summary>
        /// Audio endpoint specific.
        /// </summary>
        public readonly byte SynchAddress;

        internal UsbEndpointDescriptor()
        {
        }

        ///<summary>
        ///Returns a <see cref="T:System.String"></see> that represents the current <see cref="UsbEndpointDescriptor"></see>.
        ///</summary>
        ///
        ///<returns>
        ///A <see cref="System.String"></see> that represents the current <see cref="UsbEndpointDescriptor"></see>.
        ///</returns>
        public override string ToString()
        {
            Object[] values = {Length, DescriptorType, EndpointID, Attributes, MaxPacketSize, Interval, Refresh, SynchAddress};
            string[] names = {"Length", "DescriptorType", "EndpointID", "Attributes", "MaxPacketSize", "Interval", "Refresh", "SynchAddress"};

            return Helper.ToString("", names, ToStringParamValueSeperator, values, ToStringFieldSeperator);
        }
    }
}