// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: sourceforge.net/projects/libusbdotnet/
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Runtime.InteropServices;
using System.Threading;
using LibUsbDotNet.Main;

namespace LibUsbDotNet.Internal
{
    internal enum TransferCancelType
    {
        None,
        UserAbort,
        UserPause
    }

    internal class TransferContext
    {
        private readonly UsbEndpointBase mEndpointBase;
        private readonly SafeOverlapped mOverlapped = new SafeOverlapped();
        internal bool mAsyncCancelled = true;
        private IntPtr mBuffer;
        private GCHandle mBufferGC;
        internal int mCurrentOffset;
        internal int mCurrentRemaining;
        internal int mCurrentTransmitted;
        private int mFailRetries;
        internal int mOriginalCount;
        internal int mOriginalOffset;
        internal ManualResetEvent mOverlappedEvent = new ManualResetEvent(false);
        internal int mRequested;
        internal int mTimeout;
        private TransferCancelType mTransferCancelType = TransferCancelType.None;

        public TransferContext(UsbEndpointBase endpointBase)
        {
            mEndpointBase = endpointBase;
        }

        public SafeOverlapped Overlapped
        {
            get
            {
                return mOverlapped;
            }
        }

        public UsbEndpointBase EndpointBase
        {
            get
            {
                return mEndpointBase;
            }
        }


        internal IntPtr PtrBuf
        {
            get
            {
                return new IntPtr(mBuffer.ToInt64() + mCurrentOffset);
            }
        }

        internal int RequestCount
        {
            get
            {
                mRequested = mCurrentRemaining > UsbConstants.MAX_READ_WRITE ? UsbConstants.MAX_READ_WRITE : mCurrentRemaining;
                return mRequested;
            }
        }

        internal TransferCancelType CancelType
        {
            get
            {
                return mTransferCancelType;
            }
            set
            {
                mTransferCancelType = value;
            }
        }

        public int FailRetries
        {
            get
            {
                return mFailRetries;
            }
        }


        public void IncFailRetries()
        {
            mFailRetries++;
        }

        internal void IncrementTransfer(int amount)
        {
            lock (this)
            {
                mCurrentTransmitted += amount;
                mCurrentOffset += amount;
                mCurrentRemaining -= amount;
                mOverlappedEvent.Reset();
                mOverlapped.ClearAndSetEvent(mOverlappedEvent.SafeWaitHandle.DangerousGetHandle());
            }
        }

        internal void Setup(IntPtr buffer, int offset, int count, int timeout)
        {
            lock (this)
            {
                if (mBufferGC.IsAllocated)
                    mBufferGC.Free();
                mBuffer = buffer;

                mOriginalOffset = offset;
                mOriginalCount = count;
                mTimeout = timeout;
                Reset();
            }
        }

        internal void Setup(object buffer, int offset, int count, int timeout)
        {
            lock (this)
            {
                if (mBufferGC.IsAllocated)
                    mBufferGC.Free();
                mBufferGC = GCHandle.Alloc(buffer, GCHandleType.Pinned);
                mBuffer = mBufferGC.AddrOfPinnedObject();
                mOriginalOffset = offset;
                mOriginalCount = count;
                mTimeout = timeout;
                Reset();
            }
        }


        private void Reset()
        {
            mOverlappedEvent.Reset();
            mOverlapped.ClearAndSetEvent(mOverlappedEvent.SafeWaitHandle.DangerousGetHandle());
            mCurrentOffset = mOriginalOffset;
            mCurrentRemaining = mOriginalCount;
            mRequested = 0;
            mCurrentTransmitted = 0;
            mAsyncCancelled = true;
            EndpointBase.mCancelEvent.Reset();
            mFailRetries = 0;
        }
    }
}