﻿// Copyright © 2006-2009 Travis Robinson. All rights reserved.
// 
// website: http://sourceforge.net/projects/libusbdotnet
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Collections.Generic;
using LibUsbDotNet.Main;

namespace LibUsbDotNet.LibUsb
{
    /// <summary> 
    /// LibUsb specific members for device registry settings.  
    /// This legacy class does not actually query the windows registry as the other Registry classes do. 
    /// Instead, it wraps a <see cref="LibUsbDevice"/> and queries descriptors directly from the device 
    /// using usb IO control messages.
    /// </summary> 
    public class LegacyLibUsbRegistry : UsbRegistry
    {
        private readonly LibUsbDevice mUSBDevice;

        private LegacyLibUsbRegistry(LibUsbDevice usbDevice)
        {
            mUSBDevice = usbDevice;
            GetPropertiesSPDRP();
        }

        #region Overridden Members

        /// <summary>
        /// Opens the USB device for communucation.
        /// </summary>
        /// <returns>Return a new instance of the <see cref="UsbDevice"/> class.
        /// If the device fails to open a null refrence is return. For extended error
        /// information use the <see cref="UsbGlobals.UsbErrorEvent"/>.
        ///  </returns>
        public override UsbDevice Device
        {
            get
            {
                LibUsbDevice libUsbDevice;
                Open(out libUsbDevice);
                return libUsbDevice;
            }
        }

        /// <summary>
        /// Gets the DeviceInterfaceGuids for the WinUsb device.
        /// </summary>
        public override Guid[] DeviceInterfaceGuids
        {
            get
            {
                if (ReferenceEquals(mDeviceInterfaceGuids, null))
                {
                    if (!mDeviceProperties.ContainsKey(LIBUSB_INTERFACE_GUIDS)) return new Guid[0];

                    string[] saDeviceInterfaceGuids = mDeviceProperties[LIBUSB_INTERFACE_GUIDS] as string[];
                    if (ReferenceEquals(saDeviceInterfaceGuids, null)) return new Guid[0];

                    mDeviceInterfaceGuids = new Guid[saDeviceInterfaceGuids.Length];

                    for (int i = 0; i < saDeviceInterfaceGuids.Length; i++)
                    {
                        string sGuid = saDeviceInterfaceGuids[i].Trim(new char[] {' ', '{', '}', '[', ']', '\0'});
                        mDeviceInterfaceGuids[i] = new Guid(sGuid);
                    }
                }
                return mDeviceInterfaceGuids;
            }
        }

        /// <summary>
        /// Check this value to determine if the usb device is still connected to the bus and ready to open.
        /// </summary>
        /// <remarks>
        /// Uses the symbolic name as a unique id to determine if this device instance is still attached.
        /// </remarks>
        /// <exception cref="UsbException">An exception is thrown if the <see cref="UsbRegistry.SymbolicName"/> property is null or empty.</exception>
        public override bool IsAlive
        {
            get
            {
                if (String.IsNullOrEmpty(SymbolicName)) throw new UsbException(this, "A symbolic name is required for this property.");
                List<LegacyLibUsbRegistry> deviceList = DeviceList;
                foreach (LegacyLibUsbRegistry registry in deviceList)
                {
                    if (String.IsNullOrEmpty(registry.SymbolicName)) continue;

                    if (registry.SymbolicName == SymbolicName)
                        return true;
                }
                return false;
            }
        }

        /// <summary>
        /// Opens the USB device for communucation.
        /// </summary>
        /// <param name="usbDevice">The newly created UsbDevice.</param>
        /// <returns>True on success.</returns>
        public override bool Open(out UsbDevice usbDevice)
        {
            usbDevice = null;
            LibUsbDevice libUsbDevice;
            bool bSuccess = Open(out libUsbDevice);
            if (bSuccess)
                usbDevice = libUsbDevice;
            return bSuccess;
        }

        #endregion

        #region STATIC Members

        /// <summary>
        /// Gets a list of available LibUsb devices.
        /// </summary>
        ///
        public static List<LegacyLibUsbRegistry> DeviceList
        {
            get
            {
                List<LibUsbDevice> legacyLibUsbDeviceList = LibUsbDevice.LegacyLibUsbDeviceList;
                List<LegacyLibUsbRegistry> deviceList = new List<LegacyLibUsbRegistry>();
                foreach (LibUsbDevice usbDevice in legacyLibUsbDeviceList)
                {
                    deviceList.Add(new LegacyLibUsbRegistry(usbDevice));
                }
                return deviceList;
            }
        }

        private static string GetRegistryHardwareID(ushort vid, ushort pid, ushort rev) { return string.Format("Vid_{0}&Pid_{1}&Rev_{2}", vid.ToString("X4"), pid.ToString("X4"), rev.ToString("0000")); }

        #endregion

        /// <summary>
        /// Opens the USB device for communucation.
        /// </summary>
        /// <param name="usbDevice">The newly created UsbDevice.</param>
        /// <returns>True on success.</returns>
        public bool Open(out LibUsbDevice usbDevice)
        {
            usbDevice = null;
            bool bSuccess = mUSBDevice.Open();
            if (bSuccess)
            {
                usbDevice = mUSBDevice;
                usbDevice.mUsbRegistry = this;
            }

            return bSuccess;
        }

        private void GetPropertiesSPDRP()
        {
            mDeviceProperties.Add(DevicePropertyType.Mfg.ToString(),
                                  mUSBDevice.Info.Descriptor.ManufacturerStringIndex > 0 ? mUSBDevice.Info.ManufacturerString : string.Empty);

            mDeviceProperties.Add(DevicePropertyType.DeviceDesc.ToString(),
                                  mUSBDevice.Info.Descriptor.ProductStringIndex > 0 ? mUSBDevice.Info.ProductString : string.Empty);

            mDeviceProperties.Add("SerialNumber",
                                  mUSBDevice.Info.Descriptor.SerialStringIndex > 0 ? mUSBDevice.Info.SerialString : string.Empty);

            string fakeHardwareIds = GetRegistryHardwareID((ushort) mUSBDevice.Info.Descriptor.VendorID,
                                                           (ushort) mUSBDevice.Info.Descriptor.ProductID,
                                                           (ushort) mUSBDevice.Info.Descriptor.BcdDevice);

            mDeviceProperties.Add(DevicePropertyType.HardwareId.ToString(), new string[] {fakeHardwareIds});

            string fakeSymbolicName = fakeHardwareIds + "{00000-00000-00000-00000-00000}";

            if (mUSBDevice.Info.Descriptor.SerialStringIndex > 0)
            {
                fakeSymbolicName += "#" + mDeviceProperties["SerialNumber"] + "#";
            }
            mDeviceProperties.Add(SYMBOLIC_NAME_KEY, fakeSymbolicName);
        }
    }
}