// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: sourceforge.net/projects/libusbdotnet/
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using LibUsbDotNet.Info;
using LibUsbDotNet.Internal;
using LibUsbDotNet.Internal.LibUsb;
using LibUsbDotNet.Main;
using Microsoft.Win32.SafeHandles;

namespace LibUsbDotNet.LibUsb
{
    /// <summary> Contains members that are specific to the LibUsb-Win32 driver.
    /// </summary> 
    /// <remarks> Use the <see cref="T:LibUsbDotNet.UsbDevice"/> class instead to allow your code to work with either LibUsb or WinUsb.
    /// </remarks> 
    public class LibUsbDevice : UsbDevice
    {
        /// <summary>
        /// Kernel types supported by LibUsbDotNet.  See <see cref="KernelType"/> for more details.
        /// </summary>
        public enum LibUsbKernelType
        {
            /// <summary>
            /// LibUsb support us unavailable.
            /// </summary>
            Unknown,
            /// <summary>
            /// LibUsbDotNet native kernel driver detected.
            /// </summary>
            NativeLibUsb,
            /// <summary>
            /// Original libusb-win32 kernel driver detected.
            /// </summary>
            LegacyLibUsb
        }

        private readonly string mDeviceFilename;
        private readonly List<int> mClaimedInterfaces=new List<int>();
        private static LibUsbKernelType mLibUsbKernelType;
        private static LibUsbVersion mLibUsbKernelVersion;

        internal LibUsbDevice(UsbApiBase api, SafeHandle usbHandle, string deviceFilename)
            : base(api, usbHandle)
        {
            mDeviceFilename = deviceFilename;
        }

        /// <summary>
        /// Gets a list of libusb devices directly from the kernel; bypassing the windows registry.  
        /// This function is intended for users that do not use the native kernel driver.  
        /// If using the native kernel (sys) driver supplied with LibUsbDotNet see the <see cref="UsbGlobals.AllDevices"/>.
        /// <seealso cref="UsbGlobals"/>
        /// <seealso cref="UsbGlobals.AllLibUsbDevices"/>
        /// <seealso cref="UsbGlobals.AllWinUsbDevices"/>
        /// <seealso cref="UsbGlobals.OpenUsbDevice(LibUsbDotNet.Main.UsbDeviceFinder)"/>
        /// </summary>
        public static List<LibUsbDevice> LegacyLibUsbDeviceList
        {
            get
            {
                List<LibUsbDevice> deviceList = new List<LibUsbDevice>();
                for (int i = 1; i < UsbConstants.MAX_DEVICES; i++)
                {
                    LibUsbDevice newLibUsbDevice;
                    string deviceFileName = LibUsbDriverIO.GetDeviceNameString(i);
                    if (!Open(deviceFileName, out newLibUsbDevice)) continue;

                    newLibUsbDevice.mDeviceInfo = new UsbDeviceInfo(newLibUsbDevice);
                    newLibUsbDevice.Close();
                    deviceList.Add(newLibUsbDevice);
                }

                return deviceList;
            }
        }

        /// <summary>
        /// Closes and frees device resources.  Once closed the device cannot be reopened.  A new <see cref="UsbDevice"/> class must be obtained using the <see cref="UsbGlobals"/> class.
        /// </summary>
        /// <returns>True on success.</returns>
        public override bool Open()
        {
            if (!IsOpen)
            {
                mUsbHandle = LibUsbDriverIO.OpenDevice(mDeviceFilename);
                if (!mUsbHandle.IsClosed && !mUsbHandle.IsInvalid)
                {
                    return true;
                }
                return false;
            }
            return true;
        }

        /// <summary>
        /// Opens the USB device for communucation.
        /// </summary>
        /// <param name="deviceFilename">The LibUsb device filename to open.</param>
        /// <param name="usbDevice">The newly created UsbDevice.</param>
        /// <returns>True on success.</returns>
        public static bool Open(string deviceFilename, out LibUsbDevice usbDevice)
        {
            usbDevice = null;
            SafeFileHandle sfh = LibUsbDriverIO.OpenDevice(deviceFilename);
            if (!sfh.IsClosed && !sfh.IsInvalid)
            {
                usbDevice = new LibUsbDevice(UsbGlobals.LibUsbApi, sfh, deviceFilename);
                return true;
            }
            return false;
        }
        /// <summary>
        /// Gets the kernel driver type (sys) in use by LibUsbDotNet. This should normally return <see cref="LibUsbKernelType.NativeLibUsb"/>.
        /// </summary>
        public static LibUsbKernelType KernelType
        {
            get
            {
                if (mLibUsbKernelType == LibUsbKernelType.Unknown)
                {
                    LibUsbVersion libUsbVersion = KernelVersion;
                    mLibUsbKernelType = libUsbVersion.BcdLibUsbDotNetKernelMod != 0 ? LibUsbKernelType.NativeLibUsb : LibUsbKernelType.LegacyLibUsb;
                }

                return mLibUsbKernelType;
            }
        }

        /// <summary>
        /// Gets the kernel driver version in use by LibUsbDotNet.
        /// <alert class="note"><para>
        /// if <see cref="LibUsbVersion.BcdLibUsbDotNetKernelMod"/> is non-zero then the kernel driver is native.
        /// </para></alert>
        /// </summary>
        public static LibUsbVersion KernelVersion
        {
            get
            {
                if (mLibUsbKernelVersion.Major == 0 && mLibUsbKernelVersion.Minor == 0 && mLibUsbKernelVersion.Micro == 0 && mLibUsbKernelVersion.Nano==0)
                {
                    for (int i = 1; i < UsbConstants.MAX_DEVICES; i++)
                    {
                        LibUsbDevice newLibUsbDevice;
                        string deviceFileName = LibUsbDriverIO.GetDeviceNameString(i);
                        if (!Open(deviceFileName, out newLibUsbDevice)) continue;
                        LibUsbRequest request=new LibUsbRequest();
                        int transferred;
                        bool bSuccess = newLibUsbDevice.UsbIoSync(LibUsbIoCtl.GET_VERSION,
                                                                  request,
                                                                  LibUsbRequest.Size,
                                                                  request,
                                                                  LibUsbRequest.Size,
                                                                  out transferred);

                        newLibUsbDevice.Close();
                        if (bSuccess && transferred == LibUsbRequest.Size)
                        {
                            mLibUsbKernelVersion = request.Version;
                            break;
                        }
                    }                    
                }
                return mLibUsbKernelVersion;
            }
        }


        /// <summary>
        /// Gets the Device filename for this device.
        /// </summary>
        public string DeviceFilename
        {
            get
            {
                return mDeviceFilename;
            }
        }

        /// <summary>
        /// Claims the specified interface of the device.
        /// </summary>
        /// <param name="interfaceID">The interface to claim.</param>
        /// <returns>True on success.</returns>
        public bool ClaimInterface(int interfaceID)
        {
            if (mClaimedInterfaces.Contains(interfaceID)) return true;

            LibUsbRequest req = new LibUsbRequest();
            req.Iface.ID = interfaceID;
            req.Timeout = UsbConstants.DEFAULT_TIMEOUT;

            int ret;
            bool bSuccess = UsbIoSync(LibUsbIoCtl.CLAIM_INTERFACE, req, LibUsbRequest.Size, null, 0, out ret);
            if (bSuccess)
                mClaimedInterfaces.Add(interfaceID);

            return bSuccess;
        }

        /// <summary>
        /// Closes and frees device resources.  Once closed the device cannot be reopened.  A new <see cref="UsbDevice"/> class must be obtained using the <see cref="UsbGlobals"/> class.
        /// </summary>
        /// <returns>True on success.</returns>
        public override bool Close()
        {
            if (IsOpen)
            {
                ReleaseAllInterfaces();
                ActiveEndpoints.Clear();
                mUsbHandle.Close();
            }
            return true;
        }

        /// <summary>
        /// Releases all interface claimed by <see cref="ClaimInterface"/>.
        /// </summary>
        /// <returns>True on success.</returns>
        public int ReleaseAllInterfaces()
        {
            int claimedInterfaces = 0;
            while(mClaimedInterfaces.Count>0)
            {
                claimedInterfaces++;
                ReleaseInterface(mClaimedInterfaces[mClaimedInterfaces.Count - claimedInterfaces]);
            }

            return claimedInterfaces;
        }

         /// <summary>
        /// Releases the specified interface id that was claimed by <see cref="ClaimInterface"/>.
        /// </summary>
        /// <param name="interfaceID">The interface to release.</param>
        /// <returns>True on success.</returns>
        public bool ReleaseInterface(int interfaceID) 
        {
            LibUsbRequest req = new LibUsbRequest();
            req.Iface.ID = interfaceID;
            if (!mClaimedInterfaces.Remove(interfaceID)) return true;

            req.Timeout = UsbConstants.DEFAULT_TIMEOUT;

            int ret;
            // NOTE: A claimed interface is ALWAYS removed from the internal list.
            bool bSuccess = UsbIoSync(LibUsbIoCtl.RELEASE_INTERFACE, req, LibUsbRequest.Size, null, 0, out ret);

            return bSuccess;
        }

        /// <summary>
        /// Gets the selected configuration.
        /// </summary>
        /// <param name="config">The USB devices selected configuration index.</param>
        /// <returns>True on success.</returns>
        public override bool GetConfiguration(out byte config)

        {
            LibUsbRequest req = new LibUsbRequest();
            req.Timeout = UsbConstants.DEFAULT_TIMEOUT;

            byte[] rtnValue = new byte[1];
            int ret;
            bool bSuccess = UsbIoSync(LibUsbIoCtl.GET_CONFIGURATION, req, LibUsbRequest.Size, rtnValue, rtnValue.Length, out ret);
            if (bSuccess)
                config = rtnValue[0];
            else
                config = 0;

            return bSuccess;
        }

        /// <summary>
        /// Releases the last interface claimed by <see cref="ClaimInterface"/>.
        /// </summary>
        /// <returns>True on success.</returns>
        public bool ReleaseInterface()
        {
            //throw new UsbException(this, "You must claim an interface before releasing an interface.");
            if (mClaimedInterfaces.Count==0) return true;
            return ReleaseInterface(mClaimedInterfaces[mClaimedInterfaces.Count - 1]);
        }

        /// <summary>
        /// Sets an alternate interface for the most recent claimed interface.
        /// </summary>
        /// <param name="alternateID">The alternate interface to select for the most recent claimed interface See <see cref="ClaimInterface"/>.</param>
        /// <returns>True on success.</returns>
        public bool SetAltInterface(int alternateID)
        {
            if (mClaimedInterfaces.Count == 0) throw new UsbException(this, "You must claim an interface before setting an alternate interface.");
            return SetAltInterface(mClaimedInterfaces[mClaimedInterfaces.Count - 1], alternateID);
        }

        /// <summary>
        /// Sets an alternate interface for the specified interface.
        /// </summary>
        /// <param name="interfaceID">The interface index to specify an alternate setting for.</param>
        /// <param name="alternateID">The alternate interface setting.</param>
        /// <returns>True on success.</returns>
        public bool SetAltInterface(int interfaceID, int alternateID)
        {

            if (!mClaimedInterfaces.Contains(interfaceID)) throw new UsbException(this, String.Format("You must claim interface {0} before setting an alternate interface.", interfaceID));
            LibUsbRequest req = new LibUsbRequest();
            req.Iface.ID = interfaceID;
            req.Iface.AlternateID = alternateID;
            req.Timeout = UsbConstants.DEFAULT_TIMEOUT;

            int ret;
            return UsbIoSync(LibUsbIoCtl.SET_INTERFACE, req, LibUsbRequest.Size, null, 0, out ret);
        }

        internal bool ControlTransferEx(UsbSetupPacket setupPacket,
                             object buffer,
                             int bufferLength,
                             out int lengthTransferred,
                             int timeout)
        {
            bool bSuccess = LibUsbDriverIO.ControlTransferEx(mUsbHandle, setupPacket, buffer, bufferLength, out lengthTransferred, timeout);


            return bSuccess;
        }

        internal bool UsbIoSync(int controlCode, Object inBuffer, int inSize, object outBuffer, int outSize, out int ret)
        {
            return LibUsbDriverIO.UsbIOSync(mUsbHandle, controlCode, inBuffer, inSize, outBuffer, outSize, out ret);
        }

    }
}