// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: sourceforge.net/projects/libusbdotnet/
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using LibUsbDotNet.Internal;

namespace LibUsbDotNet.Main
{
    /// <summary> 
    /// Endpoint members common to Read, Write, Bulk, and Interrupt <see cref="T:LibUsbDotNet.Main.EndpointType"/>.
    /// </summary> 
    public class UsbEndpointBase : IDisposable
    {
        private readonly byte mEpNum;
        private readonly TransferContext mTransferContext;
        internal readonly UsbApiBase mUsbApi;
        private readonly UsbDevice mUsbDevice;
        private readonly SafeHandle mUsbHandle;
        internal ManualResetEvent mCancelEvent = new ManualResetEvent(false);
        private bool mIsDisposed;
        internal TransferDelegate mPipeTransferDelegate;

        internal UsbEndpointBase(UsbDevice usbDevice, byte epNum)
        {
            mUsbDevice = usbDevice;
            mUsbApi = mUsbDevice.mUsbApi;
            mUsbHandle = mUsbDevice.Handle;
            mEpNum = epNum;
            mTransferContext = new TransferContext(this);
            if ((mEpNum & 0x80) > 0)
            {
                mPipeTransferDelegate = ReadPipe;
            }
            else
                mPipeTransferDelegate = WritePipe;
        }

        /// <summary>
        /// Gets a value indicating if the object is disposed.
        /// </summary>
        public bool IsDisposed
        {
            get
            {
                return mIsDisposed;
            }
        }

        internal UsbDevice UsbDevice
        {
            get
            {
                return mUsbDevice;
            }
        }


        internal SafeHandle Handle
        {
            get
            {
                return mUsbHandle;
            }
        }

        /// <summary>
        /// Gets the endpoint ID for this <see cref="UsbEndpointBase"/> class.
        /// </summary>
        public byte EpNum
        {
            get
            {
                if (mIsDisposed) throw new ObjectDisposedException(GetType().Name);
                return mEpNum;
            }
        }

        internal TransferContext TransferContext
        {
            get
            {
                return mTransferContext;
            }
        }

        #region IDisposable Members

        /// <summary>
        /// Frees resources associated with the endpoint.  Once disposed this <see cref="UsbEndpointBase"/> cannot be used.
        /// </summary>
        public void Dispose()
        {
            DisposeAndRemoveFromList();
        }

        #endregion

        /// <summary>
        /// Aborts a pending IO operation on this enpoint of one exists.
        /// </summary>
        /// <returns>True on success or if no pending IO operation exits.</returns>
        public bool Abort()
        {
            if (mIsDisposed) throw new ObjectDisposedException(GetType().Name);

            lock (mTransferContext)
                return AbortPipeEx(TransferCancelType.UserAbort);
        }

        /// <summary>
        /// Discards any data that is cached in this endpoint.
        /// </summary>
        /// <returns>True on success.</returns>
        public bool Flush()
        {
            if (mIsDisposed) throw new ObjectDisposedException(GetType().Name);

            bool bSuccess = mUsbApi.FlushPipe(mUsbHandle, EpNum);

            if (!bSuccess) UsbGlobals.Error(ErrorCode.Win32Error, "FlushPipe", this);

            return bSuccess;
        }

        /// <summary>
        /// Resets the data toggle and clears the stall condition on an enpoint.
        /// </summary>
        /// <returns>True on success.</returns>
        public bool Reset()
        {
            if (mIsDisposed) throw new ObjectDisposedException(GetType().Name);

            bool bSuccess = mUsbApi.ResetPipe(mUsbHandle, EpNum);

            if (!bSuccess) UsbGlobals.Error(ErrorCode.Win32Error, "ResetPipe", this);

            return bSuccess;
        }

        /// <summary>
        /// reads/Writes data to.from this enpdoint depending on the endpoint direction bit.
        /// </summary>
        /// <param name="buffer">A caller-allocated buffer for the data that is transferred.</param>
        /// <param name="offset">Position in buffer that transferring begins.</param>
        /// <param name="length">Number of bytes, starting from thr offset parameter to transfer.</param>
        /// <param name="timeout">Maximum time to wait for the transfer to complete.</param>
        /// <param name="transferLength">Number of bytes actually transferred.</param>
        /// <returns>True on success.</returns>
        public ErrorCode Transfer(IntPtr buffer, int offset, int length, int timeout, out int transferLength)
        {
            transferLength = 0;
            ErrorCode eReturn;
            //if ((eReturn = LockEndpoint()) != ErrorCode.None) return eReturn;
            //try
            //{
                mTransferContext.Setup(buffer, offset, length, timeout);

                eReturn = DoSyncTransfer(mTransferContext, out transferLength);
            //}
            //finally
            //{
            //    UnlockEndpoint();
            //}


            return eReturn;
        }

        /// <summary>
        /// reads/Writes data to.from this enpdoint depending on the endpoint direction bit.
        /// </summary>
        /// <param name="buffer">A caller-allocated buffer for the data that is transferred.</param>
        /// <param name="offset">Position in buffer that transferring begins.</param>
        /// <param name="length">Number of bytes, starting from thr offset parameter to transfer.</param>
        /// <param name="timeout">Maximum time to wait for the transfer to complete.</param>
        /// <param name="transferLength">Number of bytes actually transferred.</param>
        /// <returns>True on success.</returns>
        public ErrorCode Transfer(object buffer, int offset, int length, int timeout, out int transferLength)
        {
            transferLength = 0;
            ErrorCode eReturn;

            mTransferContext.Setup(buffer, offset, length, timeout);

            eReturn = DoSyncTransfer(mTransferContext, out transferLength);

            return eReturn;
        }


        internal bool Abort(TransferCancelType cancelType)
        {
            return AbortPipeEx(cancelType);
        }

        internal ErrorCode DoSyncTransfer(TransferContext context, out int transferLength)
        {
            transferLength = 0;
            int iTransferred;
            ErrorCode eReturn;

            IOEndpointGlobalCancelRedo:
            do
            {
                int ret = context.EndpointBase.mPipeTransferDelegate(context.PtrBuf,
                                                                     context.RequestCount,
                                                                     out iTransferred,
                                                                     context.Overlapped.GlobalOverlapped);
                if (ret != 0 && ret != (int) UsbStatusClodes.ErrorIoPending)
                {
                    UsbError usbErr = UsbGlobals.NewUsbError(ErrorCode.Win32Error, "mPipeTransferDelegate");
                    UsbGlobals.FireUsbError(context.EndpointBase, usbErr);
                    if (!usbErr.Handled || context.FailRetries >= UsbConstants.MAX_FAIL_RETRIES_ON_HANDLED_ERROR)
                        eReturn = usbErr.ErrorCode;
                    else
                    {
                        context.IncFailRetries();
                        context.Overlapped.ClearAndSetEvent(context.mOverlappedEvent.SafeWaitHandle.DangerousGetHandle());
                        goto IOEndpointGlobalCancelRedo;
                    }
                }
                else
                {
                    if (ret == (int) UsbStatusClodes.ErrorIoPending)
                        context.mAsyncCancelled = false;

                    eReturn = WaitForAsyncTransferResult(context, out iTransferred);

                    if (eReturn != ErrorCode.None)
                    {
                        if (eReturn == ErrorCode.IoEndpointGlobalCancelRedo)
                        {
                            context.Overlapped.ClearAndSetEvent(context.mOverlappedEvent.SafeWaitHandle.DangerousGetHandle());
                            goto IOEndpointGlobalCancelRedo;
                        }
                        break;
                    }
                    transferLength = context.mCurrentTransmitted;
                }
            } while (eReturn == ErrorCode.None && iTransferred == UsbConstants.MAX_READ_WRITE && context.mCurrentRemaining > 0);

            return eReturn;
        }

        private bool AbortPipeEx(TransferCancelType cancelType)
        {
            mTransferContext.mAsyncCancelled = true;
            mTransferContext.CancelType = cancelType;
            mCancelEvent.Set();
            bool bSuccess = mUsbApi.AbortPipe(mUsbHandle, EpNum);

            if (!bSuccess) UsbGlobals.Error(ErrorCode.Win32Error, "AbortPipe", this);

            return bSuccess;
        }

        private void DisposeAndRemoveFromList()
        {
            if (!mIsDisposed)
            {
                UsbEndpointReader epReader = this as UsbEndpointReader;
                if (!ReferenceEquals(epReader, null))
                {
                    if (epReader.DataReceivedEnabled) epReader.DataReceivedEnabled = false;
                }
                Abort();
                mUsbDevice.ActiveEndpoints.RemoveFromList(this);
            }
            mIsDisposed = true;
        }

        private int ReadPipe(IntPtr pBuffer, int bufferLength, out int lengthTransferred, IntPtr pOverlapped)
        {
            bool bSuccess = mUsbApi.ReadPipe(Handle, EpNum, pBuffer, bufferLength, out lengthTransferred, pOverlapped);
            if (!bSuccess) return Marshal.GetLastWin32Error();
            return 0;
        }

        private ErrorCode WaitForAsyncTransferResult(TransferContext context, out int ret)
        {
            ret = 0;
            int iWait =
                WaitHandle.WaitAny(
                    new WaitHandle[]
                        {context.mOverlappedEvent, context.EndpointBase.mCancelEvent},
                    context.mTimeout,
                    false);
            if (iWait == 0)
            {
                bool bSuccess = mUsbApi.GetOverlappedResult(context.EndpointBase.Handle, context.Overlapped.GlobalOverlapped, out ret, true);
                if (!bSuccess)
                {
                    UsbError usbErr = UsbGlobals.NewUsbError(ErrorCode.Win32Error, "GetOverlappedResult");
                    UsbGlobals.FireUsbError(context.EndpointBase, usbErr);
                    if (!usbErr.Handled || context.FailRetries >= UsbConstants.MAX_FAIL_RETRIES_ON_HANDLED_ERROR)
                        return usbErr.ErrorCode;

                    context.IncFailRetries();
                    return ErrorCode.IoEndpointGlobalCancelRedo;
                }
                context.IncrementTransfer(ret);
                return ErrorCode.None;
            }
            switch (iWait)
            {
                case WaitHandle.WaitTimeout:
                    context.EndpointBase.Abort();
                    return ErrorCode.IoTimedOut;
                case 1:
                    if (context.CancelType == TransferCancelType.UserPause)
                    {
                        context.CancelType = TransferCancelType.None;
                        context.EndpointBase.mCancelEvent.Reset();
                        return ErrorCode.IoEndpointGlobalCancelRedo;
                    }
                    return ErrorCode.IoCancelled;
                case 2:
                    context.EndpointBase.Abort();
                    context.CancelType = TransferCancelType.None;
                    Debug.Print("IoEndpointGlobalCancelRedo:True");
                    return ErrorCode.IoEndpointGlobalCancelRedo;
            }

            return ErrorCode.None;
        }

        private int WritePipe(IntPtr pBuffer, int bufferLength, out int lengthTransferred, IntPtr pOverlapped)
        {
            bool bSuccess = mUsbApi.WritePipe(Handle, EpNum, pBuffer, bufferLength, out lengthTransferred, pOverlapped);
            if (!bSuccess) return Marshal.GetLastWin32Error();
            return 0;
        }

        #region Nested Types

        internal delegate int TransferDelegate(IntPtr pBuffer, int bufferLength, out int lengthTransferred, IntPtr pOverlapped);

        #endregion
    }
}