// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: sourceforge.net/projects/libusbdotnet/
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Runtime.InteropServices;
using LibUsbDotNet.Internal;

namespace LibUsbDotNet.Main
{
    /// <summary> Describes a Usb error or setup API error.
    /// </summary> 
    public class UsbError : EventArgs
    {
        /// <summary>
        /// The the error is <see cref="Handled"/> field is set to true for errors resulting from endpoint read/write errors.  The operation will retry instead of exiting with an error code. 
        /// </summary>
        public bool Handled;

        internal string mDescription;

        internal ErrorCode mErrorCode;
        internal int mWin32ErrorNumber;
        internal string mWin32ErrorString = "None";

        internal UsbError(ErrorCode errorCode, string description)
        {
            mErrorCode = errorCode;
            if (errorCode == ErrorCode.Win32Error)
            {
                mWin32ErrorNumber = Marshal.GetLastWin32Error();
                mWin32ErrorString = Kernel32.FormatSystemMessage(mWin32ErrorNumber);
            }
            mDescription = description;
        }

        internal UsbError(ErrorCode errorCode, int win32ErrorNumber, string win32ErrorString, string description)
        {
            mErrorCode = errorCode;
            mWin32ErrorNumber = win32ErrorNumber;
            mWin32ErrorString = win32ErrorString;
            mDescription = description;
        }

        /// <summary>
        /// Gets the general errorcode.
        /// </summary>
        public ErrorCode ErrorCode
        {
            get
            {
                return mErrorCode;
            }
        }

        /// <summary>
        /// Gets the Windows specific error number.  Only valid when <see cref="ErrorCode"/> is set to <see cref="Main.ErrorCode"/>.<see cref="Main.ErrorCode.Win32Error"/>.
        /// </summary>
        public int Win32ErrorNumber
        {
            get
            {
                return mWin32ErrorNumber;
            }
        }

        /// <summary>
        /// Gets the general description for the error.
        /// </summary>
        public string Description
        {
            get
            {
                return mDescription;
            }
        }

        /// <summary>
        /// Gets the Windows specific error string. Only valid when <see cref="ErrorCode"/> is set to <see cref="Main.ErrorCode"/>.<see cref="Main.ErrorCode.Win32Error"/>.
        /// </summary>
        public string Win32ErrorString
        {
            get
            {
                return mWin32ErrorString;
            }
        }


        ///<summary>
        ///Returns a <see cref="T:System.String"></see> that represents the current <see cref="UsbError"></see>.
        ///</summary>
        ///
        ///<returns>
        ///A <see cref="System.String"></see> that represents the current <see cref="UsbError"></see>.
        ///</returns>
        public override string ToString()
        {
            if (Win32ErrorNumber != 0)
            {
                return string.Format("{0}:{1}\r\n{2}:{3}", ErrorCode, Description, Win32ErrorNumber, mWin32ErrorString);
            }
            return string.Format("{0}:{1}", ErrorCode, Description);
        }
    }
}