// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: http://sourceforge.net/projects/libusbdotnet
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Runtime.InteropServices;
using System.Text;
using LibUsbDotNet.Descriptors;
using LibUsbDotNet.Info;
using LibUsbDotNet.Internal;
using LibUsbDotNet.LibUsb;
using LibUsbDotNet.Main;
using LibUsbDotNet.WinUsb;

namespace LibUsbDotNet
{
    /// <summary>Contains non-driver specific USB device communication members.</summary>
    /// <remarks>This class is compatible with both WinUsb and LibUsb-Win32 device drivers in a generic fashion, 
    /// IE: Your application code can operate with complete disregard of the driver type (LibUsb or WinUsb).</remarks>
    public abstract class UsbDevice
    {
        //private const int DEF_DEVICELOCK_TIMEOUT = 1000;

        #region REQUIRED Override Members

        /// <summary>
        /// Closes and frees device resources.  Once closed the device cannot be reopened.  A new <see cref="UsbDevice"/> class must be obtained using the <see cref="UsbGlobals"/> class.
        /// </summary>
        /// <returns>True on success.</returns>
        public abstract bool Close();
        ///<summary>
        /// Opens/re-opens this USB device instance for communication.
        ///</summary>
        ///<returns>True if the device is allready opened or was opened successfully.  False if the device does not exists or is no longer valid.</returns>
        public abstract bool Open();
        #endregion

        private readonly UsbEndpointList mActiveEndpoints;
        internal readonly UsbApiBase mUsbApi;
        private List<UsbConfigInfo> mConfigs;
        internal UsbDeviceInfo mDeviceInfo;

        internal SafeHandle mUsbHandle;
        internal UsbRegistry mUsbRegistry;

        internal UsbDevice(UsbApiBase usbApi, SafeHandle usbHandle)
        {
            mUsbApi = usbApi;
            mUsbHandle = usbHandle;
            mActiveEndpoints = new UsbEndpointList();
        }

        #region OPTIONAL Override Members

        /// <summary>
        /// Gets the selected configuration.
        /// </summary>
        /// <param name="config">The USB devices selected configuration index.</param>
        /// <returns>True on success.</returns>
        public virtual bool GetConfiguration(out byte config)
        {
            config = 0;
            byte[] buf = new byte[1];
            int uTransferLength;

            UsbSetupPacket setupPkt = new UsbSetupPacket();
            setupPkt.RequestType = UsbRequestType.EndpointIn | UsbRequestType.TypeStandard | UsbRequestType.RecipDevice;
            setupPkt.Request = DeviceRequestType.GetConfiguration;
            setupPkt.Value = 0;
            setupPkt.Index = 0;
            setupPkt.Length = 1;

            bool bSuccess = ControlTransfer(ref setupPkt, buf, buf.Length, out uTransferLength);
            if (bSuccess && uTransferLength == 1)
            {
                config = buf[0];
                return true;
            }
            UsbGlobals.Error(ErrorCode.Win32Error, "GetConfiguration", this);
            return false;
        }

        #endregion

        #region STATIC Members

        internal static bool EndpointExists(IEnumerable<UsbConfigInfo> configList, byte endpointAddress)
        {
            foreach (UsbConfigInfo configInfo in configList)
            {
                foreach (UsbInterfaceInfo interfaceInfo in configInfo.InterfaceInfoList)
                {
                    foreach (UsbEndpointInfo endpointInfo in interfaceInfo.EndpointInfoList)
                    {
                        if (endpointInfo.Descriptor.EndpointID == endpointAddress) return true;
                    }
                }
            }
            return false;
        }

        internal static List<UsbConfigInfo> GetDeviceConfigs(UsbDevice usbDevice)
        {
            List<UsbConfigInfo> rtnConfigs = new List<UsbConfigInfo>();

            byte[] cfgBuffer = new byte[UsbConstants.MAX_CONFIG_SIZE];

            int iConfigs = usbDevice.Info.Descriptor.ConfigurationCount;
            for (int iConfig = 0; iConfig < iConfigs; iConfig++)
            {
                int iBytesTransmitted;
                bool bSuccess = usbDevice.GetDescriptor((byte) DescriptorType.Configuration, 0, 0, cfgBuffer, cfgBuffer.Length, out iBytesTransmitted);
                if (bSuccess)
                {
                    if (iBytesTransmitted >= UsbConfigDescriptor.Size && cfgBuffer[1] == (byte) DescriptorType.Configuration)
                    {
                        UsbConfigDescriptor configDescriptor = new UsbConfigDescriptor();
                        Helper.BytesToObject(cfgBuffer, 0, Math.Min(UsbConfigDescriptor.Size, cfgBuffer[0]), configDescriptor);

                        if (configDescriptor.TotalLength == iBytesTransmitted)
                        {
                            List<byte[]> rawDescriptorList = new List<byte[]>();
                            int iRawLengthPosition = configDescriptor.Length;
                            while (iRawLengthPosition < configDescriptor.TotalLength)
                            {
                                byte[] rawDescriptor = new byte[cfgBuffer[iRawLengthPosition]];
                                if (iRawLengthPosition + rawDescriptor.Length > iBytesTransmitted)
                                    throw new UsbException(usbDevice, "Descriptor length is out of range.");

                                Array.Copy(cfgBuffer, iRawLengthPosition, rawDescriptor, 0, rawDescriptor.Length);
                                rawDescriptorList.Add(rawDescriptor);
                                iRawLengthPosition += rawDescriptor.Length;
                            }
                            rtnConfigs.Add(new UsbConfigInfo(usbDevice, configDescriptor, ref rawDescriptorList));
                        }
                        else
                            UsbGlobals.Error(ErrorCode.InvalidConfig,
                                             "GetDeviceConfigs: USB config descriptor length doesn't match the length received.",
                                             usbDevice);
                    }
                    else
                        UsbGlobals.Error(ErrorCode.InvalidConfig, "GetDeviceConfigs: USB config descriptor is invalid.", usbDevice);
                }
                else
                    UsbGlobals.Error(ErrorCode.InvalidConfig, "GetDeviceConfigs", usbDevice);
            }
            return rtnConfigs;
        }

        #endregion

        /// <summary>
        /// Gets a value indication if the device handle is valid.
        /// </summary>
        public bool IsOpen
        {
            get
            {
                return ((mUsbHandle != null) && !mUsbHandle.IsClosed) && !mUsbHandle.IsInvalid;
            }
        }

        /// <summary>
        /// A list of endpoints that have beened opened by this <see cref="UsbDevice"/> class.
        /// </summary>
        public UsbEndpointList ActiveEndpoints
        {
            get { return mActiveEndpoints; }
        }

        ///<summary>
        /// Gets the available configurations for this <see cref="UsbDevice"/>
        ///</summary>
        /// <remarks>
        /// The first time this property is accessed it will query the <see cref="UsbDevice"/> for all configurations.  Subsequent request will return a cached copy of all configurations.
        /// </remarks>
        public ReadOnlyCollection<UsbConfigInfo> Configs
        {
            get
            {
                if (ReferenceEquals(mConfigs, null))
                {
                    mConfigs = GetDeviceConfigs(this);
                }
                return mConfigs.AsReadOnly();
            }
        }


        internal SafeHandle Handle
        {
            get { return mUsbHandle; }
        }

        /// <summary>
        /// Gets the actual device descriptor the the current <see cref="UsbDevice"/>.
        /// </summary>
        public UsbDeviceInfo Info
        {
            get
            {
                if (ReferenceEquals(mDeviceInfo, null))
                {
                    mDeviceInfo = new UsbDeviceInfo(this);
                }
                return mDeviceInfo;
            }
        }

        /// <summary>
        /// Gets the <see cref="UsbRegistry"/> class that opened the device, or null if the device was not opened by the <see cref="UsbRegistry"/> class.
        /// </summary>
        public UsbRegistry UsbRegistryInfo
        {
            get { return mUsbRegistry; }
        }


        /// <summary>
        /// Sends/Receives an IO control message to endpoint 0.
        /// </summary>
        /// <param name="setupPacket">Contains parameters for the control request. See section 9.3 USB Device Requests of the Universal Serial Bus Specification Revision 2.0 for more information. </param>
        /// <param name="buffer">Data to be sent/received from the device.</param>
        /// <param name="bufferLength">Length of the buffer param.</param>
        /// <param name="lengthTransferred">Number of bytes sent or received (depends on the direction of the control transfer).</param>
        /// <returns>True on success.</returns>
        public bool ControlTransfer(ref UsbSetupPacket setupPacket, object buffer, int bufferLength, out int lengthTransferred)
        {
            bool bSuccess = mUsbApi.ControlTransfer(mUsbHandle, setupPacket, buffer, bufferLength, out lengthTransferred);

            if (!bSuccess)
                UsbGlobals.Error(ErrorCode.Win32Error, "ControlTransfer", this);

            return bSuccess;
        }

        /// <summary>
        /// Gets a specific descriptor from the device. See <see cref="DescriptorType"/> for more information.
        /// </summary>
        /// <param name="descriptorType">The descriptor type ID to retrieve; this is usually one of the <see cref="DescriptorType"/> enumerations.</param>
        /// <param name="index">Descriptor index.</param>
        /// <param name="langId">Descriptor language id.</param>
        /// <param name="buffer">Memory to store the returned descriptor in.</param>
        /// <param name="bufferLength">Length of the buffer parameter in bytes.</param>
        /// <param name="transferLength">The number of bytes transferred to buffer upon success.</param>
        /// <returns>True on success.</returns>
        public bool GetDescriptor(byte descriptorType, byte index, short langId, object buffer, int bufferLength, out int transferLength)
        {
            transferLength = 0;

            bool wasOpen = IsOpen;
            if (!wasOpen) Open();
            if (!IsOpen) return false;

            bool bSuccess = mUsbApi.GetDescriptor(mUsbHandle, descriptorType, index, (ushort) langId, buffer, bufferLength, out transferLength);

            if (!bSuccess)
                UsbGlobals.Error(ErrorCode.Win32Error, "GetDescriptor", this);

            if (!wasOpen && IsOpen) Close();

            return bSuccess;
        }


        /// <summary>
        /// Asking for the zero'th index is special - it returns a string
        /// descriptor that contains all the language IDs supported by the
        /// device. Typically there aren't many - often only one. The
        /// language IDs are 16 bit numbers, and they start at the third byte
        /// in the descriptor. See USB 2.0 specification, section 9.6.7, for
        /// more information on this. 
        /// </summary>
        /// <returns>A collection of LCIDs that the current <see cref="UsbDevice"/> supports.</returns>
        public bool GetLangIDs(out short[] langIDs)
        {
            LangStringDescriptor sd = new LangStringDescriptor();

            int ret;
            bool bSuccess = GetDescriptor((byte) DescriptorType.String, 0, 0, sd, LangStringDescriptor.Size, out ret);
            if (bSuccess)
            {
                if (ret == sd.Length)
                {
                    ret -= 2; // dscType, dscSize
                    langIDs = new short[ret/2];
                    for (int i = 0; i < langIDs.Length; i++)
                        langIDs[i] = sd.LangIds[i];
                    return true;
                }
            }
            langIDs = null;
            UsbGlobals.Error(ErrorCode.Win32Error, "GetLangIDs", this);
            return false;
        }

        /// <summary>
        /// Gets a string descriptor from the device.
        /// </summary>
        /// <param name="stringData">Buffer to store the returned string in upon success.</param>
        /// <param name="langId">The language ID to retrieve the string in. (0x409 for english).</param>
        /// <param name="stringIndex">The string index to retrieve.</param>
        /// <returns>True on success.</returns>
        public bool GetString(out string stringData, short langId, byte stringIndex)
        {
            stringData = null;
            int iTransferLength;
            StringDescriptor sd = new StringDescriptor();
            bool bSuccess = GetDescriptor((byte) DescriptorType.String, stringIndex, langId, sd, StringDescriptor.Size, out iTransferLength);

            if (bSuccess && iTransferLength > UsbDescriptor.Size)
                stringData = Encoding.Unicode.GetString(sd.Str, 0, iTransferLength - UsbDescriptor.Size).TrimEnd(new char[] {'\0'});
            else if (!bSuccess)
                UsbGlobals.Error(ErrorCode.Win32Error, "GetString:GetDescriptor", this);
            else
                stringData = String.Empty;

            return bSuccess;
        }


        /// <summary>
        /// Opens an endpoint for reading
        /// </summary>
        /// <param name="readEndpointID">Endpoint number for read operations.</param>
        /// <param name="readBufferSize">Size of the read buffer allocated for the <see cref="UsbEndpointReader.DataReceived"/> event.</param>
        /// <returns>A <see cref="UsbEndpointReader"/> class ready for reading.
        /// If the specified endpoint has allready been opened, the original <see cref="UsbEndpointReader"/> object will be returned.
        /// </returns>
        /// <exception cref="UsbException">If the endpoint does not exist.</exception>
        public UsbEndpointReader OpenEndpointReader(ReadEndpointID readEndpointID, int readBufferSize)
        {
            if (!EndpointExists(Configs, (byte) readEndpointID))
                throw new UsbException(this,
                                       String.Format("Endpoint {0} (0x{1}) does not exists.", readEndpointID, ((byte) readEndpointID).ToString("X2")));
            UsbEndpointReader epNew = new UsbEndpointReader(this, readBufferSize, readEndpointID);
            return (UsbEndpointReader) mActiveEndpoints.Add(epNew);
        }


        /// <summary>
        /// Opens an endpoint for reading
        /// </summary>
        /// <param name="readEndpointID">Endpoint number for read operations.</param>
        /// <returns>A <see cref="UsbEndpointReader"/> class ready for reading.
        /// If the specified endpoint has allready been opened, the original <see cref="UsbEndpointReader"/> object will be returned.
        /// </returns>
        /// <exception cref="UsbException">If the endpoint does not exist.</exception>
        public UsbEndpointReader OpenEndpointReader(ReadEndpointID readEndpointID) { return OpenEndpointReader(readEndpointID, UsbEndpointReader.DefReadBufferSize); }

        /// <summary>
        /// Opens an endpoint for writing
        /// </summary>
        /// <param name="writeEndpointID">Endpoint number for read operations.</param>
        /// <returns>A <see cref="UsbEndpointWriter"/> class ready for writing.
        /// If the specified endpoint has allready been opened, the original <see cref="UsbEndpointWriter"/> object will be returned.
        /// </returns>
        /// <exception cref="UsbException">If the endpoint does not exist.</exception>
        public UsbEndpointWriter OpenEndpointWriter(WriteEndpointID writeEndpointID)
        {
            if (!EndpointExists(Configs, (byte) writeEndpointID))
                throw new UsbException(this,
                                       String.Format("Endpoint {0} (0x{1}) does not exists.", writeEndpointID, ((byte) writeEndpointID).ToString("X2")));
            UsbEndpointWriter epNew = new UsbEndpointWriter(this, writeEndpointID);
            return (UsbEndpointWriter) mActiveEndpoints.Add(epNew);
        }

        /// <summary>
        /// Sets the USB devices configuration profile. 
        /// </summary>
        /// <param name="config">The configuration index to select.</param>
        /// <returns>True on success.</returns>
        public bool SetConfiguration(byte config)
        {
            int uTransferLength;

            UsbSetupPacket setupPkt = new UsbSetupPacket();
            setupPkt.RequestType = UsbRequestType.EndpointOut | UsbRequestType.TypeStandard | UsbRequestType.RecipDevice;
            setupPkt.Request = DeviceRequestType.SetConfiguration;
            setupPkt.Value = config;
            setupPkt.Index = 0;
            setupPkt.Length = 0;

            bool bSuccess = ControlTransfer(ref setupPkt, null, 0, out uTransferLength);
            if (!bSuccess)
                UsbGlobals.Error(ErrorCode.Win32Error, "SetConfiguration", this);

            return bSuccess;
        }

        /// <summary>
        /// Gets the selected alternate interface of the specified interface.
        /// </summary>
        /// <param name="interfaceID">The interface settings number (index) to retrieve the selected alternate interface setting for.</param>
        /// <param name="selectedAltInterfaceID">The alternate interface setting selected for use with the specified interface.</param>
        /// <returns>True on success.</returns>
        public bool GetAltInterfaceSetting(byte interfaceID, out byte selectedAltInterfaceID)
        {
            byte[] buf = new byte[1];
            int uTransferLength;

            UsbSetupPacket setupPkt = new UsbSetupPacket();
            setupPkt.RequestType = UsbRequestType.EndpointIn | UsbRequestType.TypeStandard | UsbRequestType.RecipInterface;
            setupPkt.Request = DeviceRequestType.GetInterface;
            setupPkt.Value = 0;
            setupPkt.Index = interfaceID;
            setupPkt.Length = 1;

            bool bSuccess = ControlTransfer(ref setupPkt, buf, buf.Length, out uTransferLength);
            if (bSuccess && uTransferLength == 1)
                selectedAltInterfaceID = buf[0];
            else
                selectedAltInterfaceID = 0;

            return bSuccess;
        }

        /// <summary>
        /// Kernel driver modes supported by LibUsbDotNet.
        /// </summary>
        public enum DriverModeType
        {
            /// <summary>
            /// Using LibUsb kernel driver (Legacy or Native).
            /// </summary>
            LibUsb,
            /// <summary>
            /// Using WinUsb user-mode driver.
            /// </summary>
            WinUsb
        }
        /// <summary>
        /// Returns the DriverMode this USB device is using.
        /// </summary>
        public DriverModeType DriverMode
        {
            get
            {
                if (this is LibUsbDevice) return DriverModeType.LibUsb;
                if (this is WinUsbDevice) return DriverModeType.WinUsb;

                throw new UsbException(this, "Unknown DriverModeType:" + GetType().FullName);
            }
        }
    }
}