// Copyright  2006-2009 Travis Robinson. All rights reserved.
// 
// website: http://sourceforge.net/projects/libusbdotnet
// e-mail:  libusbdotnet@gmail.com
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. or 
// visit www.gnu.org.
// 
// 
using System;
using System.Collections.Generic;
using System.IO;
using LibUsbDotNet.Internal.LibUsb;
using LibUsbDotNet.Internal.WinUsb;
using LibUsbDotNet.LibUsb;
using LibUsbDotNet.Main;
using LibUsbDotNet.WinUsb;
using Debug=System.Diagnostics.Debug;

namespace LibUsbDotNet
{
    /// <summary> 
    /// Static class for opening, enumerating and finding USB devices.  
    /// </summary> 
    public static class UsbGlobals
    {
        private const string LIBUSB_SYS = "libusb0.sys";
        private const string WINUSB_DLL = "winusb.dll";

        private static int _lastErrorNumber;
        private static string _lastErrorString = String.Empty;
        private static LibUsbAPI _libUsbApi;
        private static WinUsbAPI _winUsbApi;
        private static bool ForceLegacyLibUsb = false;
        private static object mHasWinUsbDriver = null;
        private static object mHasLibUsbDriver = null;

        /// <summary>
        /// True if the WinUsb driver is found on the system.
        /// </summary>
        public static bool HasWinUsbDriver
        {
            get
            {
                if (mHasWinUsbDriver == null)
                {
                    mHasWinUsbDriver = false;
                    DirectoryInfo diSysDriver = new DirectoryInfo(Environment.SystemDirectory);
                    if (diSysDriver.Exists)
                    {
                        FileInfo[] fiLibusbs = diSysDriver.GetFiles(WINUSB_DLL);
                        if (fiLibusbs.Length > 0)
                            mHasWinUsbDriver = true;
                    }
                }
                return (bool)mHasWinUsbDriver;
            }
        }

        /// <summary>
        /// True if the LibUsb driver is found on the system.
        /// </summary>
        public static bool HasLibUsbDriver
        {
            get
            {
                if (mHasLibUsbDriver == null)
                {
                    mHasLibUsbDriver = false;
                    string sysDriver = Path.Combine(Environment.SystemDirectory, "drivers");
                    DirectoryInfo diSysDriver = new DirectoryInfo(sysDriver);
                    if (diSysDriver.Exists)
                    {
                        FileInfo[] libUsbSysFileInfo = diSysDriver.GetFiles(LIBUSB_SYS);
                        if (libUsbSysFileInfo.Length > 0)
                            mHasLibUsbDriver = true;
                    }
                }
                return (bool)mHasLibUsbDriver;
            }
        }

        /// <summary>
        /// Gets a list of all available USB devices (WinUsb and/or LibUsb).
        /// </summary>
        public static UsbRegDeviceList AllDevices
        {
            get
            {
                UsbRegDeviceList regDevReturnList = new UsbRegDeviceList();

                UsbRegDeviceList winUsbList = AllWinUsbDevices;
                foreach (UsbRegistry winUsbRegistry in winUsbList)
                    regDevReturnList.Add(winUsbRegistry);


                UsbRegDeviceList libUsbList = AllLibUsbDevices;
                foreach (UsbRegistry libUsbRegistry in libUsbList)
                    regDevReturnList.Add(libUsbRegistry);

                return regDevReturnList;
            }
        }


        /// <summary>
        /// Gets a new <see cref="UsbRegDeviceList"/> instance containing all device <see cref="UsbRegistry"/> classes that will be controlled by Microsofts WinUsb driver.
        /// </summary>
        public static UsbRegDeviceList AllWinUsbDevices
        {
            get
            {
                UsbRegDeviceList regDevList = new UsbRegDeviceList();
                if (HasWinUsbDriver)
                {
                    List<WinUsbRegistry> winUsbRegistry = WinUsbRegistry.DeviceList;
                    foreach (WinUsbRegistry usbRegistry in winUsbRegistry)
                        regDevList.Add(usbRegistry);
                }
                return regDevList;
            }
        }

        /// <summary>
        /// Gets a new <see cref="UsbRegDeviceList"/> instance containing all device <see cref="UsbRegistry"/> classes that will be controlled by Stephan Meyers libusb-win32 driver.
        /// </summary>
        public static UsbRegDeviceList AllLibUsbDevices
        {
            get
            {
                UsbRegDeviceList regDevList = new UsbRegDeviceList();

                if (HasLibUsbDriver && !ForceLegacyLibUsb && LibUsbDevice.KernelType == LibUsbDevice.LibUsbKernelType.NativeLibUsb)
                {
                    List<LibUsbRegistry> libUsbRegistry = LibUsbRegistry.DeviceList;
                    foreach (LibUsbRegistry usbRegistry in libUsbRegistry)
                        regDevList.Add(usbRegistry);
                }
                else if (HasLibUsbDriver)
                {
                    List<LegacyLibUsbRegistry> libUsbRegistry = LegacyLibUsbRegistry.DeviceList;
                    foreach (LegacyLibUsbRegistry usbRegistry in libUsbRegistry)
                        regDevList.Add(usbRegistry);
                }
                return regDevList;
            }
        }

        internal static WinUsbAPI WinUsbApi
        {
            get
            {
                if (ReferenceEquals(_winUsbApi, null))
                    _winUsbApi = new WinUsbAPI();
                return _winUsbApi;
            }
        }

        internal static LibUsbAPI LibUsbApi
        {
            get
            {
                if (ReferenceEquals(_libUsbApi, null))
                    _libUsbApi = new LibUsbAPI();
                return _libUsbApi;
            }
        }

        /// <summary>
        /// Returns the last error string reported by LibUsbDotNet.
        /// </summary>
        public static string LastErrorString
        {
            get
            {
                return _lastErrorString;
            }
        }

        /// <summary>
        /// Returns the last error number reported by LibUsbDotNet.
        /// </summary>
        public static int LastErrorNumber
        {
            get
            {
                return _lastErrorNumber;
            }
        }

        /// <summary>
        /// Global static error event for all Usb errors.
        /// </summary>
        /// <example>
        /// Sample code to reset an endpoint if a critical error occurs.
        /// <code>
        /// // Hook the usb error handler function
        /// UsbGlobals.UsbErrorEvent += UsbErrorEvent;
        ///private void UsbErrorEvent(object sender, UsbError e)
        ///{
        /// // If the error is from a usb endpoint
        /// if (sender is UsbEndpointBase)
        /// {
        ///     // If the endpoint transfer failed
        ///     if (e.Win32ErrorNumber == 31)
        ///     {
        ///         // If the USB device is still open, connected, and valid
        ///         if (usb.IsOpen)
        ///         {
        ///             // Try to reset then endpoint
        ///             if (((UsbEndpointBase) sender).Reset())
        ///             {
        ///                 // Endpoint reset successful.
        ///                 // Tell LibUsbDotNet to ignore this error and continue.
        ///                 e.Handled = true;
        ///             }
        ///         }
        ///     }
        /// }
        /// }
        /// </code>
        /// </example>
        public static event EventHandler<UsbError> UsbErrorEvent;

        /// <summary>
        /// Opens a WinUsb device by its DeviceInterfaceGUID.
        /// </summary>
        /// <remarks>
        /// This is the Microsoft-recommended way for opening a WinUsb device.  
        /// LibUsb device can be opened in this way as well.  In order to open
        /// LibUsb devices in this manner, an entry must be added to the driver
        /// inf file:
        /// <para>[Install.HW]</para>
        /// <para>Addreg=Add_LibUsb_Guid_Reg</para>
        /// <para>[Add_LibUsb_Guid_Reg]</para>
        /// <para>HKR,,LibUsbInterfaceGUIDs,0x10000,"{Your-Unique-Guid-String}"</para>
        /// </remarks>
        /// <param name="devInterfaceGuid">Device Interface GUID of the usb device to open.</param>
        /// <param name="usbDevice">On success, a new <see cref="UsbDevice"/> instance.</param>
        /// <returns>True on success.</returns>
        public static bool OpenUsbDevice(ref Guid devInterfaceGuid, out UsbDevice usbDevice)
        {
            usbDevice = null;
            UsbRegDeviceList usbRegDevices = AllDevices;
            foreach (UsbRegistry usbRegistry in usbRegDevices)
            {
                foreach (Guid guid in usbRegistry.DeviceInterfaceGuids)
                {
                    if (guid == devInterfaceGuid)
                    {
                        usbDevice = usbRegistry.Device;
                        if (usbDevice != null) return true;
                    }
                }
            }

            return false;
        }

        /// <summary>
        /// Opens the usb device that matches the find predicate.
        /// </summary>
        /// <param name="findDevicePredicate">The predicate function used to find the usb device.</param>
        /// <returns>An valid/open usb device class if the device was found or Null if the device was not found.</returns>
        public static UsbDevice OpenUsbDevice(Predicate<UsbRegistry> findDevicePredicate)
        {
            UsbDevice usbDeviceFound;

            UsbRegDeviceList allDevices = AllDevices;
            UsbRegistry regDeviceFound = allDevices.Find(findDevicePredicate);

            if (ReferenceEquals(regDeviceFound, null)) return null;

            usbDeviceFound = regDeviceFound.Device;

            return usbDeviceFound;
        }

        /// <summary>
        /// Opens the usb device that matches the <see cref="UsbDeviceFinder"/>.
        /// </summary>
        /// <param name="usbDeviceFinder">The <see cref="UsbDeviceFinder"/> class used to find the usb device.</param>
        /// <returns>An valid/open usb device class if the device was found or Null if the device was not found.</returns>
        public static UsbDevice OpenUsbDevice(UsbDeviceFinder usbDeviceFinder)
        {
            return OpenUsbDevice(usbDeviceFinder.Check);
        }

        internal static void Error(ErrorCode errorCode, string description)
        {
            Error(errorCode, description, "Global");
        }


        internal static ErrorCode Error(ErrorCode errorCode, string description, object sender)
        {
            UsbError usbErr = NewUsbError(errorCode, description);

            if (!FireUsbError(sender, usbErr))
            {
                Debug.Print(_lastErrorString);
            }
            return errorCode;
        }

        internal static bool FireUsbError(object sender, UsbError usbError)
        {
            EventHandler<UsbError> temp = UsbErrorEvent;
            if (!ReferenceEquals(null, temp))
            {
                temp(sender, usbError);
                return true;
            }
            return false;
        }

        internal static UsbError NewUsbError(ErrorCode errorCode, string description)
        {
            UsbError usbErr = new UsbError(errorCode, description);
            lock (_lastErrorString)
            {
                _lastErrorNumber = usbErr.mWin32ErrorNumber;
                _lastErrorString = usbErr.ToString();
            }
            return usbErr;
        }
    }
}